import abc
import functools
import time
import inspect
from typing import Optional, Any, TypeVar, Type, Callable

from amino import Map, List, Empty, Just, __, L, Maybe, _, Lists

from ribosome.record import any_field, dfield, list_field, field, RecordMeta, Record, bool_field

_machine_attr = '_machine'
_message_attr = '_message'
_prio_attr = '_prio'
_dyn_attr = '_dyn'
default_prio = 0.5
fallback_prio = 0.3
override_prio = 0.8


def _field_namespace(fields, opt_fields, varargs):
    namespace = Map()
    for fname in fields:
        namespace[fname] = any_field()
    for fname, val in opt_fields:
        namespace[fname] = any_field(initial=val)
    if varargs:
        namespace[varargs] = list_field()
    return namespace


def _init_field_metadata(inst):
    def set_missing(name, default):
        if not hasattr(inst, name):
            setattr(inst, name, default)
    List(
        ('_field_varargs', None),
        ('_field_order', []),
        ('_field_count_min', 0),
    ).map2(set_missing)


def _update_field_metadata(inst, fields, opt_fields, varargs):
    if varargs is not None:
        inst._field_varargs = varargs
    inst._field_order += list(fields) + List(*opt_fields).map(__[0])
    inst._field_count_min += len(fields)
    inst._field_count_max = (
        Empty() if inst._field_varargs
        else Just(inst._field_count_min + len(opt_fields)))


class MessageMeta(RecordMeta, abc.ABCMeta):

    def __new__(cls, name, bases, namespace, fields=[], opt_fields=[], varargs=None, skip_fields=False, **kw):
        ''' create a subclass of PRecord
        **fields** is a list of strings used as names of mandatory PRecord fields
        **opt_fields** is a list of (string, default) used as fields with initial values
        the order of the names is preserved in **_field_order**.
        **varargs** is an optional field name where unmatched args are stored.
        **skip_fields** indicates that the current class is a base class (like Message). If those classes were processed
        here, all their subclasses would share the metadata, and get all fields set by other subclasses.
        **_field_count_min** and **_field_count_max** are used by `MessageCommand`
        '''
        ns = Map() if skip_fields else _field_namespace(fields, opt_fields, varargs)
        inst = super().__new__(cls, name, bases, ns ** namespace, **kw)
        if not skip_fields:
            _init_field_metadata(inst)
            _update_field_metadata(inst, fields, opt_fields, varargs)
        return inst

    def __init__(cls, name, bases, namespace, **kw):
        super().__init__(name, bases, namespace)

M = TypeVar('M', bound='Message')


@functools.total_ordering
class Message(Record, metaclass=MessageMeta, skip_fields=True):
    ''' Interface between vim commands and state.
    Provides a constructor that allows specification of fields via positional arguments.
    '''
    time = field(float)
    prio = dfield(0.5)
    bang = bool_field()
    range = any_field(initial=0)

    def __new__(cls, *args, **kw):
        count = len(cls._field_order)
        sargs, vargs = args[:count], args[count:]
        vmap = Map({cls._field_varargs: vargs}) if cls._field_varargs else Map()
        field_map = vmap ** Map(zip(cls._field_order, sargs))
        ext_kw = field_map ** kw + ('time', time.time())
        return super().__new__(cls, **ext_kw)

    def __init__(self, *a: Any, **kw: Any) -> None:
        pass

    @classmethod
    def from_msg(cls: Type[M], other: 'Message') -> Callable[..., M]:
        return lambda *a, **kw: cls(*a, bang=other.bang, range=other.range, **kw)

    @property
    def pub(self):
        return Publish(self)

    def __lt__(self, other):
        if isinstance(other, Message):
            if self.prio == other.prio:
                return self.time < self.time
            else:
                return self.prio < other.prio
        else:
            return True

    def at(self, prio):
        return self.set(prio=float(prio))

    @property
    def _str_extra(self) -> List[Any]:
        return Lists.wrap(self._field_order) // L(Maybe.getattr)(self, _) + self.bang.l('bang!')


def message_definition_module() -> Optional[str]:
    return inspect.currentframe().f_back.f_back.f_globals['__name__']


def message(name, *fields, mod=None, **kw):
    module = mod or message_definition_module()
    return MessageMeta(name, (Message,), dict(__module__=module), fields=fields, **kw)


def json_message(name, *fields, mod=None, **kw):
    opt = (('options', Map()),)
    module = mod or message_definition_module()
    return message(name, *fields, opt_fields=opt, mod=module, **kw)


class Publish(Message, fields=('message',)):

    def __str__(self):
        return 'Publish({})'.format(str(self.message))

__all__ = ('message', 'Message', 'json_message', 'Publish')



import time
import uuid
import inspect
from typing import Callable, TypeVar, Type, Any, Generic
import asyncio

import toolz

from amino import Maybe, _, List, Map, Empty, L, __, Just, Either, Lists, Left, Right, IO
from amino.util.string import camelcaseify
from amino.io import IO
from amino.lazy import lazy
from amino.func import flip

from ribosome.machine.message_base import Message
from ribosome.logging import print_ribo_log_info
from ribosome.data import Data
from ribosome.nvim import NvimIO, NvimFacade
from ribosome.machine.interface import MachineI
from ribosome.machine.transition import (Handler, TransitionResult, may_handle, Error, Debug, handle, _io_result,
                                         _recover_error, DynHandler)
from ribosome.machine.message_base import _machine_attr
from ribosome.request.command import StateCommand
from ribosome.process import NvimProcessExecutor
from ribosome.machine.messages import (RunNvimIO, RunIO, UnitIO, RunCorosParallel, SubProcessSync, RunIOsParallel,
                                       ShowLogInfo, Nop, RunIOAlg, TransitionException, Info, RunNvimIOAlg, DataIO,
                                       RunNvimUnitIO)
from ribosome.machine.handler import Handlers, DynHandlerJob, AlgHandlerJob, HandlerJob
from ribosome.machine import trans
from ribosome.machine.trans import Propagate

A = TypeVar('A')
D = TypeVar('D', bound=Data)


def nio(f: Callable[[NvimFacade], Any]) -> RunNvimIO:
    return RunNvimIO(NvimIO(f))


def unit_nio(f: Callable[[NvimFacade], None]) -> RunNvimUnitIO:
    return RunNvimUnitIO(NvimIO(f))


class MachineBase(Generic[D], MachineI):
    _data_type = Data

    def __init__(self, parent: 'Machine'=None, title=None) -> None:
        self.parent = Maybe(parent)
        self._title = Maybe(title)
        self.uuid = uuid.uuid4()
        self._reports = List()
        self._min_report_time = 0.1

    @property
    def title(self):
        return self._title.or_else(
            List.wrap(type(self).__module__.rsplit('.')).reversed.lift(1)
        ) | 'machine'

    @lazy
    def _message_handlers(self):
        def create(prio, h):
            h = List.wrap(h).apzip(_.message).map2(flip)
            return prio, Handlers(prio, Map(h))
        return Map(toolz.groupby(_.prio, self._handlers)).map(create)

    @property
    def _handlers(self):
        methods = inspect.getmembers(type(self), lambda a: hasattr(a, _machine_attr))
        return List.wrap(methods).map2(L(Handler.create)(self, _, _))

    @property
    def _default_handler(self):
        return DynHandler(self, 'unhandled', type(self).unhandled, None, 0, True)

    @property
    def prios(self):
        return self._message_handlers.k

    def handler_job_type(self, handler: Handler) -> Type[HandlerJob]:
        return DynHandlerJob if handler.dyn else AlgHandlerJob

    def process(self, data: D, msg, prio=None) -> TransitionResult:
        self.prepare(msg)
        handler = self._resolve_handler(msg, prio)
        job = self.handler_job_type(handler)(self, data, msg, handler, self._data_type)
        result = job.run()
        self._check_time(job.start_time, msg)
        return result

    def prepare(self, msg):
        pass

    def loop_process(self, data, msg, prio=None):
        def loop(z, m) -> None:
            self.parent % __.log_message(m, self.title)
            return z.accum(self.loop_process(z.data, m))
        return self.process(data, msg, prio).fold(loop)

    def log_message(self, msg: Message, name: str) -> None:
        pass

    def _resolve_handler(self, msg, prio):
        f = __.handler(msg)
        return (
            self._message_handlers.v.find_map(f)
            if prio is None else
            (self._message_handlers.get(prio) // f)
        ) | (lambda: self._default_handler)

    def unhandled(self, data: D, msg: Message):
        return Just(TransitionResult.unhandled(data))

    def _command_by_message_name(self, name: str):
        msg_name = camelcaseify(name)
        return (
            self._message_handlers.to_list
            .sort_by(__[0])
            .map2(lambda a, b: b.handlers)
            .find_map(__.find_key(lambda a: a.__name__ in [name, msg_name]))
        )

    def command(self, name: str, args: list):
        return (
            self._command_by_message_name(name)
            .map(lambda a: StateCommand(a[0]))
            .map(__.dispatch(self, args))
            .or_else(L(self._invalid_command)(name))
        )

    def _invalid_command(self, name):
        self.log.error('plugin "{}" has no command "{}"'.format(self.title, name))
        return Empty()

    @may_handle(RunNvimUnitIO)
    def message_run_nvim_unit_io(self, data: D, msg: RunNvimUnitIO):
        msg.io.unsafe_perform_io(self.vim)

    def _run_nio(self, io: RunNvimIOAlg) -> Either[str, List[Message]]:
        result = io.attempt(self.vim)
        if result.value is None:
            self.log.error(f'NvimIO returned None: {io}')
        return result

    @trans.multi(RunNvimIO, trans.e)
    def message_run_nvim_io(self, data: D, msg: RunNvimIO) -> Either[str, List[Message]]:
        return self._run_nio(msg.io)

    @trans.relay(RunNvimIOAlg)
    def message_run_nvim_io_alg(self, data: D, msg: RunNvimIOAlg) -> Either[str, List[Message]]:
        return Propagate.from_either(self._run_nio(msg.io))

    def _run_io(self, io: IO[A]) -> Either[str, List[Message]]:
        result = io.attempt
        if result.value is None:
            self.log.error(f'IO returned None: {io}')
        return _io_result(result)

    @handle(RunIO)
    def message_run_io(self, data: D, msg):
        return self._run_io(msg.io)

    @trans.relay(RunIOAlg)
    def message_run_io_alg(self, data: D, msg: RunIOAlg):
        return self._run_io(msg.io)

    @handle(UnitIO)
    def message_run_unit_io(self, data: D, msg):
        return self._run_io(msg.io.replace(Just(Nop())))

    @handle(DataIO)
    def message_data_io(self, data: D, msg):
        return self._run_io(msg.cons(IO.now(data)))

    @may_handle(RunCorosParallel)
    def message_run_coros_parallel(self, data: D, msg: RunCorosParallel) -> Message:
        async def wrap() -> Either[str, List[Message]]:
            try:
                results = await asyncio.gather(*msg.coros)
            except Exception as e:
                self.log.caught_exception('running coroutines', e)
                return Left(f'running coros failed: {e}')
            else:
                return Lists.wrap(results).traverse(L(_recover_error)('parallel coros', _), Maybe)
        return wrap()

    @may_handle(SubProcessSync)
    async def message_sub_process_sync(self, data: D, msg: SubProcessSync) -> Message:
        executor = NvimProcessExecutor(self.vim)
        await executor.run(msg.job)
        return msg.result(msg.job.result)

    @may_handle(RunIOsParallel)
    def message_run_ios_parallel(self, data: D, msg: RunIOsParallel) -> Message:
        coros = msg.ios / _.coro
        async def wrap() -> Maybe[List[Message]]:
            try:
                results = await asyncio.gather(*coros)
            except Exception as e:
                self.log.caught_exception('running coroutines', e)
                return Left(TransitionException('running coroutines', e))
            else:
                return Right(
                    Lists.wrap(results) /
                    __.value_or(lambda a: TransitionException('running IOs in parallel', a.cause).pub)
                )
        return wrap()

    @may_handle(Error)
    def message_error(self, data: D, msg: Error) -> None:
        self.log.error(msg)

    @may_handle(Info)
    def message_info(self, data: D, msg: Info) -> None:
        self.log.info(msg.message)

    @may_handle(TransitionException)
    def transition_exception(self, data, msg):
        self.log.caught_exception(msg.context, msg.exc)

    @may_handle(Debug)
    def message_debug(self, data, msg):
        self.log.debug(msg.message)

    @may_handle(ShowLogInfo)
    def show_log_info(self, data: D, msg: ShowLogInfo) -> Message:
        print_ribo_log_info(self.log.verbose)

    def bubble(self, msg):
        self.parent.cata(_.bubble, lambda: self.send)(msg)

    def _check_time(self, start_time, msg):
        dur = time.time() - start_time
        self.log.debug1(self._format_report, msg, dur)
        if dur > self._min_report_time:
            self._reports = self._reports.cat((msg, dur))

    def report(self):
        if self._reports:
            self.log.info('time-consuming messages in {}:'.format(self.title))
            self._reports.map2(self._format_report) % self.log.info

    def _format_report(self, msg, dur):
        return '{} took {:.4f}s for {} to process'.format(msg, dur, self.title)

__all__ = ('MachineBase',)



from ribosome.nvim import NvimFacade
from ribosome.request import command, msg_command, json_msg_command, function, msg_function, json_msg_function
from ribosome.machine import (StateMachine, handle, may_handle, PluginStateMachine, RootMachine, ModularMachine,
                              Transitions)
from ribosome.process import ProcessExecutor, Job, NvimProcessExecutor
from ribosome.plugin import NvimStatePlugin, NvimPlugin

in_vim = True

__all__ = ('Plugin', 'command', 'NvimFacade', 'StateMachine', 'development', 'NvimStatePlugin', 'msg_command',
           'ProcessExecutor', 'Job', 'NvimPlugin', 'PluginStateMachine', 'in_vim', 'json_msg_command', 'RootMachine',
           'handle', 'may_handle', 'ModularMachine', 'Transitions', 'function', 'msg_function', 'json_msg_function',
           'NvimProcessExecutor')



from kallikrein import Expectation, kf, k
from kallikrein.matchers.either import be_right

from amino.test.path import fixture_path
from amino import Either, Right

from ribosome.rpc import define_handlers, rpc_handlers
from ribosome.test.integration.klk import PluginIntegrationKlkSpec, later

from integration._support.plugin import TestPluginUnlooped


class HostSpec(PluginIntegrationKlkSpec):
    '''start a host from its class file $start_host
    '''

    @property
    def _prefix(self) -> str:
        return 'host'

    @property
    def autostart_plugin(self) -> bool:
        return False

    @property
    def plugin_class(self) -> Either[str, type]:
        return Right(TestPluginUnlooped)

    def start_host(self) -> Expectation:
        exe = fixture_path('host', 'run')
        plug = self.rplugin_path.get_or_raise
        channel = self.vim.call(
            'jobstart',
            ['python3', str(exe), str(plug)],
            dict(rpc=True)
        ).get_or_raise
        handlers = rpc_handlers(self.plugin_class.get_or_raise)
        define_handlers(channel, handlers, 'host', str(plug)).attempt(self.vim).get_or_raise
        self._wait(.2)
        self.command_exists('Go')
        return later(kf(self.cmd_sync, 'Go').must(be_right))

__all__ = ('HostSpec',)



from amino import Map

from typing import Callable, Any

from ribosome.record import Record, field


class Data(Record):
    sub_states = field(Map, initial=Map())

    def sub_state_m(self, name):
        return self.sub_states.get(name)

    def sub_state(self, name, default: Callable[[], Any]):
        return self.sub_state_m(name) | default

    def with_sub_state(self, name, state):
        new_states = self.sub_states + (name, state)
        return self.set(sub_states=new_states)

__all__ = ('Data',)



import re
import abc
from typing import Callable

from amino import Maybe, Either, __, Right, L, List, _, Map, Try

from ribosome.logging import Logging
from ribosome.nvim import NvimFacade


class VimCallback(Logging, metaclass=abc.ABCMeta):

    def __init__(self, target) -> None:
        self.target = target
        self.vim = self.target.root

    @abc.abstractmethod
    def __call__(self, data):
        ...


class SpecialCallback:

    def __init__(self, name) -> None:
        self.name = name

    def __str__(self):
        return self.name


class CallbackSpec(Logging, metaclass=abc.ABCMeta):

    def __init__(self, spec) -> None:
        self.spec = spec

    @abc.abstractmethod
    def func(self, vim: NvimFacade) -> Either[str, Callable]:
        ...

    def __call__(self, vim: NvimFacade, *a):
        return self.func(vim) // __(*a)

    def __str__(self):
        return '{}({})'.format(self.__class__.__name__, self.spec)


class PythonCallbackSpecBase(CallbackSpec):

    def _inst(self, vim, name):
        return (Try(name, vim)
                if isinstance(name, type) and issubclass(name, VimCallback)
                else Right(name))

    @abc.abstractmethod
    def _func(self):
        ...

    def func(self, vim):
        return self._func // L(self._inst)(vim, _)


class PythonCallbackSpec(PythonCallbackSpecBase):

    @property
    def _func(self):
        return Either.import_path(self.spec)


class StrictPythonCallbackSpec(PythonCallbackSpecBase):

    @property
    def _func(self):
        return Right(self.spec)


class VimFuncCallbackSpec(CallbackSpec):

    def func(self, vim):
        return Right(L(vim.call)(self.spec))


class VarCallbackSpec(CallbackSpec):

    def func(self, vim):
        return Right(L(vim.vars)(self.spec))

_py_callback_re = re.compile('^py:(.+)')
_vim_callback_re = re.compile('^vim:(.+)')
_var_callback_re = re.compile('^var:(.+)')
_special_callback_re = re.compile('^s:(.+)')

_callback_res = List(
    (_py_callback_re, PythonCallbackSpec),
    (_vim_callback_re, VimFuncCallbackSpec),
    (_var_callback_re, VarCallbackSpec),
)


def _cb_err(data):
    return 'invalid callback string: {}'.format(data)


def parse_callback(data: str, rex, tpe: type):
    return (
        Maybe(rex.match(data)) /
        __.group(1) /
        tpe
    )


def parse_special_callback(data, special: Map) -> Maybe[CallbackSpec]:
    return (
        Maybe(_special_callback_re.match(data)) /
        __.group(1) //
        special.get /
        StrictPythonCallbackSpec
    )


def parse_callback_spec(data: str, special=Map()) -> Maybe[CallbackSpec]:
    spec = special.empty.no.flat_m(L(parse_special_callback)(data, special))
    other = lambda: _callback_res.flat_map2(L(parse_callback)(data, _, _)).head
    return spec.o(other).to_either(_cb_err(data))

__all__ = ('VimCallback', 'parse_callback_spec')



import sys
import inspect
import functools
from typing import TypeVar, Type, Callable, Tuple

from neovim.msgpack_rpc import MsgpackStream, AsyncSession, Session
from neovim.plugin import Host
from neovim.api import Nvim
from neovim.msgpack_rpc.event_loop.base import BaseEventLoop
from neovim.msgpack_rpc.event_loop.asyncio import AsyncioEventLoop
from neovim.msgpack_rpc.event_loop.uv import UvEventLoop

from amino import Either, _, L, do, Path, Maybe, Lists, Right, Try, Logger

from ribosome.nvim import NvimFacade
from ribosome.logging import ribo_log, Logging
from ribosome.rpc import rpc_handler_functions, RpcHandlerFunction
from ribosome import NvimPlugin

Loop = TypeVar('Loop', bound=BaseEventLoop)
NP = TypeVar('NP', bound=NvimPlugin)


class PluginHost(Host, Logging):

    def _load(self, data: Tuple[str, Type[NP]]) -> None:
        file, tpe = data
        instance = tpe(self._configure_nvim_for(tpe))
        handlers = rpc_handler_functions(instance)
        handlers % L(self.register)(file, _)
        self._specs[file] = handlers / _.spec.encode

    def register(self, file: str, handler: RpcHandlerFunction) -> None:
        spec = handler.spec
        method = spec.rpc_method(file)
        func = functools.partial(self._wrap_function, handler.func, spec.sync, True, None, method)
        self._copy_attributes(handler.func, func)
        target = self._request_handlers if spec.sync else self._notification_handlers
        target[method] = func


def session(Loop: Loop, *args: str, transport_type='stdio', **kwargs: str) -> Session:
    return Session(AsyncSession(MsgpackStream(Loop(transport_type, *args, **kwargs))))


def host(Loop: Loop) -> PluginHost:
    sess = session(Loop)
    nvim = Nvim.from_session(sess)
    return PluginHost(nvim)


def start_host(data: Tuple[str, Type[NP]], uv_loop: bool) -> None:
    Loop = UvEventLoop if uv_loop else AsyncioEventLoop
    host(Loop).start(data)


@do
def cls_from_file(file: str) -> Either[str, Tuple[str, Type[NP]]]:
    mod = yield Either.import_file(Path(file))
    all = yield Maybe.getattr(mod, '__all__').to_either(f'file `{file}` does not define `__all__`')
    name = yield Lists.wrap(all).head.to_either(f'`{mod}.__all__` is empty')
    tpe = yield Maybe.getattr(mod, name).to_either(f'{mod}.{name} does not exist')
    yield Right((file, tpe))


@do
def cls_from_path(path: str) -> Either[str, Tuple[str, Type[NP]]]:
    tpe = yield Either.import_path(path)
    file = yield Try(inspect.getfile, tpe)
    yield Right((file, tpe))


def log() -> Logger:
    NvimFacade.stdio_with_logging('ribosome_start_host')
    return ribo_log


def start(desc: str, target: str, cls: Callable[[str], Either[str, Tuple[str, Type[NP]]]]) -> None:
    def error(msg: str) -> None:
        log().error(f'invalid rplugin {desc} `{target}`: {msg}')
    try:
        cls(target).map(L(start_host)(_, True)).leffect(error)
    except Exception as e:
        log().caught_exception_error(f'starting host with {desc} `{target}`', e)


def start_file(file: str) -> None:
    start('file', file, cls_from_file)


def start_cls(cls: str) -> None:
    start('class', cls, cls_from_path)


def cli() -> None:
    start_file(sys.argv[1])


__all__ = ('start_host', 'cli', 'start_file', 'start_cls')



import abc
from typing import Union, Any, Callable, Type

import neovim

from amino import List

from ribosome.nvim import NvimFacade
from ribosome.machine import StateMachine, Message
from ribosome.logging import nvim_logging, Logging
from ribosome.request import msg_command, msg_function, command, function, json_msg_command
from ribosome.machine.base import ShowLogInfo
from ribosome.machine.scratch import Mapping
from ribosome.rpc import rpc_handlers_json
from ribosome.record import encode_json
from ribosome.machine.messages import UpdateState


class NvimPlugin(Logging):
    name: str = None
    prefix: str = None

    def __init_subclass__(cls: type, name: str=None, prefix: str=None, debug=False) -> None:
        if name is not None:
            setup_plugin(cls, name, prefix or name, debug)

    def __init__(self, nvim: Union[NvimFacade, neovim.Nvim]) -> None:
        self.vim = NvimFacade(nvim, self.plugin_name) if isinstance(nvim, neovim.Nvim) else nvim
        self.setup_logging()

    def setup_logging(self) -> None:
        self.file_log_handler = nvim_logging(self.vim)

    @property
    def loop(self) -> Any:
        return self.vim.loop

    @property
    def plugin_name(self) -> str:
        return type(self).name or 'ribosome'

    @abc.abstractmethod
    def stage_1(self) -> None:
        ...

    def stage_2(self) -> None:
        pass

    def stage_3(self) -> None:
        pass

    def stage_4(self) -> None:
        pass

    def rpc_handlers(self) -> List[dict]:
        return rpc_handlers_json(type(self))

    def set_log_level(self, level: str) -> None:
        self.file_log_handler.setLevel(level)

    def plug_command(self, plug_name: str, cmd_name: str, *args: str) -> None:
        self.state().plug_command(plug_name, cmd_name, args)


class NvimStatePlugin(NvimPlugin):

    def __init_subclass__(cls: type, name: str=None, prefix: str=None, debug: bool=False) -> None:
        super().__init_subclass__(name, prefix, debug)
        if cls.name:
            setup_state_plugin(cls, cls.name, cls.prefix, debug)

    @abc.abstractmethod
    def state(self) -> StateMachine:
        ...

    def message_log(self) -> List[Message]:
        return self.state().message_log // encode_json

    def state_data(self) -> str:
        return self.state().data.json | '{}'


UnitF = Callable[..., None]


class Helpers:

    def __init__(self, cls: type, name: str, prefix: str) -> None:
        self.cls = cls
        self.name = name
        self.prefix = prefix

    def _handler(self, name: str, dec: Callable[..., UnitF], handler: UnitF, *a: Any, **kw: Any) -> None:
        func = dec(*a, name=name, **kw)(handler)
        setattr(self.cls, name, func)

    def name_handler(self, suf: str, dec: Callable[..., UnitF], handler: UnitF, *a: Any, **kw: Any) -> None:
        n = f'{self.name}_{suf}'
        self._handler(n, dec, handler, *a, **kw)

    def handler(self, suf: str, dec: Callable[..., UnitF], handler: UnitF, *a: Any, **kw: Any) -> None:
        n = f'{self.prefix}_{suf}'
        self._handler(n, dec, handler, *a, **kw)

    def msg_cmd(self, suf: str, msg: type) -> None:
        self.handler(suf, msg_command, lambda: None, msg)

    def msg_fun(self, suf: str, msg: type) -> None:
        self.handler(suf, msg_function, lambda: None, msg)

    def json_msg_cmd(self, suf: str, msg: type) -> None:
        self.handler(suf, json_msg_command, lambda: None, msg)


def setup_plugin(cls: Type[NvimPlugin], name: str, prefix: str, debug: bool) -> None:
    help = Helpers(cls, name, prefix)
    cls.name = name
    cls.prefix = prefix
    cls.debug = debug
    help.msg_cmd('show_log_info', ShowLogInfo)
    help.handler('log_level', command, cls.set_log_level)
    help.msg_fun('mapping', Mapping)
    help.name_handler('stage_1', command, cls.stage_1, sync=True)
    help.name_handler('stage_2', command, cls.stage_2, sync=True)
    help.name_handler('stage_3', command, cls.stage_3, sync=True)
    help.name_handler('stage_4', command, cls.stage_4, sync=True)
    help.name_handler('rpc_handlers', function, cls.rpc_handlers, sync=True)
    help.handler('plug', command, cls.plug_command)


def setup_state_plugin(cls: Type[NvimStatePlugin], name: str, prefix: str, debug: bool) -> None:
    help = Helpers(cls, name, prefix)
    help.handler('state', function, cls.state_data)
    help.json_msg_cmd('update_state', UpdateState)
    if debug:
        help.name_handler('message_log', function, cls.message_log, sync=True)

__all__ = ('NvimPlugin', 'NvimStatePlugin')



from amino import Right, Either

from kallikrein import Expectation, unsafe_k
from kallikrein.matchers.either import be_right
from kallikrein.matchers.maybe import be_just

from integration._support.plugin import TestPluginLooped, TestPlugin, TestPluginUnlooped
from integration._support.base import IntegrationSpecBase

specs = '''\
plugin with {}
simple commands producing output $startup
launch a scratch buffer $scratch
execute coroutines in parallel $coros_parallel
execute IOs in parallel $ios_parallel
'''


class _PluginSpecBase(IntegrationSpecBase):

    def _last_output(self, content):
        return self._log_line(-1, be_just(content))

    def _startup(self) -> Expectation:
        i = 99932
        fun_text = 'message function test'
        self.vim.cmd_sync('Go')
        val = TestPlugin.test_value.format(i)
        unsafe_k(self.vim.call('Value', i)).must(be_right(val))
        self._last_output(TestPlugin.test_go)
        val2 = TestPlugin.test_fun.format(i)
        unsafe_k(self.vim.call('Fun', i)).must(be_right(val2))
        self.vim.call('MsgFun', fun_text)
        self._last_output(fun_text)
        self.vim.cmd_sync('Err')
        self._last_output(TestPlugin.test_error)
        self.vim.cmd_sync('St')
        self._wait(1)
        return self._last_output('6')

    def _scratch(self) -> Expectation:
        self.vim.cmd_sync('Go')
        self.vim.cmd_sync('Scratch')
        self._window_count(2)
        self.vim.cmd_sync('CheckScratch')
        self._last_output('1')
        self.vim.cmd_sync('ScratchTest')
        self._last_output(TestPlugin.test_scratch)
        self.vim.window.close()
        self._window_count(1)
        self.vim.cmd_sync('CheckScratch')
        return self._last_output('0')

    def _coros_parallel(self) -> Expectation:
        self.vim.cmd_sync('Go')
        self.vim.cmd_sync('RunParallel')
        self._log_contains('sleeping in 2')
        self._log_contains('sleeping in 1')
        return self._log_contains('sleeping in 0')

    def _ios_parallel(self) -> Expectation:
        self.vim.cmd_sync('Go')
        self.vim.cmd_sync('RunParallelIOs')
        self._log_contains('sleeping in 2')
        self._log_contains('sleeping in 1')
        return self._log_contains('sleeping in 0')


class LoopedPluginSpec(_PluginSpecBase):
    __doc__ = specs.format('permanent main loop thread')

    @property
    def plugin_class(self) -> Either[str, type]:
        return Right(TestPluginLooped)

    def startup(self) -> Expectation:
        return self._startup()

    def scratch(self) -> Expectation:
        return self._scratch()

    def coros_parallel(self) -> Expectation:
        return self._coros_parallel()

    def ios_parallel(self) -> Expectation:
        return self._ios_parallel()


class UnloopedPluginSpec(_PluginSpecBase):
    __doc__ = specs.format('on-demand main loop')

    @property
    def plugin_class(self) -> Either[str, type]:
        return Right(TestPluginUnlooped)

    def startup(self) -> Expectation:
        return self._startup()

    def scratch(self) -> Expectation:
        return self._scratch()

    def coros_parallel(self) -> Expectation:
        return self._coros_parallel()

    def ios_parallel(self) -> Expectation:
        return self._ios_parallel()

__all__ = ('LoopedPluginSpec', 'UnloopedPluginSpec')



from ribosome.machine.base import MachineBase
from ribosome.machine.state import StateMachine, PluginStateMachine, RootMachine
from ribosome.machine.transition import TransitionFailed, handle, may_handle, either_msg, either_handle, Error
from ribosome.machine.message_base import Message, message, json_message
from ribosome.machine.messages import Nop, Quit, Info
from ribosome.machine.modular import ModularMachine
from ribosome.machine.transitions import Transitions

__all__ = ('MachineBase', 'StateMachine', 'PluginStateMachine', 'handle', 'may_handle', 'TransitionFailed',
           'RootMachine', 'ModularMachine', 'Nop', 'either_msg', 'either_handle', 'Transitions', 'Message', 'Error',
           'message', 'json_message', 'Quit', 'Info')



import functools

import neovim

from ribosome.request.base import RequestHandler, MessageRequestHandler, JsonMessageRequestHandler


class Command(RequestHandler):

    @property
    def desc(self):
        return 'command'

    @property
    def neovim_cmd(self):
        @neovim.command(self.name, nargs=self.nargs, **self.kw)
        @functools.wraps(self._fun)
        def neovim_cmd_wrapper(obj, *rpc_args):
            return self.dispatch(obj, rpc_args)
        return neovim_cmd_wrapper


class MessageCommand(Command, MessageRequestHandler):
    pass


class JsonMessageCommand(Command, JsonMessageRequestHandler):
    pass


class StateCommand(MessageCommand):

    def __init__(self, msg: type, **kw) -> None:
        def fun():
            pass
        super(StateCommand, self).__init__(fun, msg, **kw)

    def _call_fun(self, obj, *args, **kw):
        return self._message(*args, **kw)

    @property
    def _infer_name(self):
        return self._message.__name__


def command(**kw):
    def neovim_cmd_decorator(fun):
        handler = Command(fun, **kw)
        return handler.neovim_cmd
    return neovim_cmd_decorator


def msg_command(msg: type, **kw):
    def neovim_msg_cmd_decorator(fun):
        handler = MessageCommand(fun, msg, **kw)
        return handler.neovim_cmd
    return neovim_msg_cmd_decorator


def json_msg_command(msg: type, **kw):
    def neovim_json_msg_cmd_decorator(fun):
        handler = JsonMessageCommand(fun, msg, **kw)
        return handler.neovim_cmd
    return neovim_json_msg_cmd_decorator


__all__ = ('command', 'msg_command', 'json_msg_command')



import re
import abc
from typing import Callable

from ribosome.machine import message, handle, may_handle
from ribosome.nvim import HasNvim, NvimFacade, ScratchBuffer
from ribosome.machine.state import KillMachine, SubMachine
from ribosome.machine.base import UnitIO

from amino import Map, Boolean, __, Empty
from amino.io import IO
from amino.lazy import lazy

Mapping = message('Mapping', 'uuid', 'keyseq')
Quit = message('Quit')


class ScratchMachine(SubMachine, HasNvim, metaclass=abc.ABCMeta):

    def __init__(self, vim: NvimFacade, scratch: ScratchBuffer, parent=None, title=None) -> None:
        self.scratch = scratch
        SubMachine.__init__(self, parent, title=title)
        HasNvim.__init__(self, vim)
        self._create_mappings()
        self._create_autocmds()

    @abc.abstractproperty
    def prefix(self) -> str:
        ...

    @abc.abstractproperty
    def mappings(self) -> Map[str, Callable]:
        ...

    @lazy
    def _quit_seq(self):
        return '%%quit%%'

    @lazy
    def _internal_mappings(self):
        return Map({
            self._quit_seq: Quit,
        })

    def _create_mappings(self):
        self.mappings.k / self._create_mapping

    def _create_mapping(self, keyseq, to=Empty()):
        m = re.match('%(.*)%', keyseq)
        ks = '<{}>'.format(m.group(1)) if m else keyseq
        toseq = to | keyseq
        cmd = self._mapping_call(toseq)
        self.scratch.buffer.nmap(ks, ':{}<cr>'.format(cmd))

    def _mapping_call(self, seq):
        return 'call {}Mapping(\'{}\', \'{}\')'.format(self.prefix, self.uuid, seq)

    def _create_autocmds(self):
        cmd = self.scratch.buffer.autocmd('BufWipeout', self._mapping_call(self._quit_seq))
        cmd.run_async()

    @handle(Mapping)
    def input(self, data, msg):
        return (
            Boolean(str(msg.uuid) == str(self.uuid))
            .flat_maybe(self.mappings.get(msg.keyseq))
            .or_else(self._internal_mappings.get(msg.keyseq)) /
            __()
        )

    @may_handle(Quit)
    def quit(self, data, msg):
        close = IO.delay(self.scratch.close)
        return UnitIO(close), KillMachine(self.uuid).pub

__all__ = ('ScratchMachine',)



from ribosome import command, function, NvimStatePlugin

from amino import List
from ribosome.logging import Logging


class FacadeTestPlugin(NvimStatePlugin, Logging):

    @function()
    def all_vars(self):
        return self.vim.buffer.vars.all

    @command(sync=True)
    def go(self):
        pass

__all__ = ('FacadeTestPlugin',)



import time
from datetime import datetime
from typing import Callable, Any, Union, Type, TypeVar

from amino.test.spec import default_timeout
from amino import List, __

from kallikrein import Expectation, kf
from kallikrein.expectable import Expectable
from kallikrein.matcher import BoundMatcher
from kallikrein.matchers.length import have_length
from kallikrein.matchers.comparison import greater
from kallikrein.matchers import contain, equal
from kallikrein.matchers.lines import have_lines
from kallikrein.matchers.maybe import be_just
from kallikrein.matchers.either import be_right
from kallikrein.matchers.typed import have_type

from ribosome.test.integration.spec import (VimIntegrationSpecI, VimIntegrationSpec, ExternalIntegrationSpec,
                                            PluginIntegrationSpec)
from ribosome.nvim.components import Buffer
from ribosome.machine import Message


def later_f(exp: Callable[[], Expectation], timeout: float=None, intval: float=0.1) -> Expectation:
    to = default_timeout if timeout is None else timeout
    start = datetime.now()
    while (not exp().unsafe_eval and (datetime.now() - start).total_seconds() < to):
        time.sleep(intval)
    exp().fatal_eval()
    return exp()


def later(exp: Expectation, timeout: float=None, intval: float=0.1) -> None:
    return later_f(lambda: exp, timeout, intval)


M = TypeVar('M', bound=Message)


class VimIntegrationKlkHelpers(VimIntegrationSpecI):

    def _wait_for(self, pred: Callable[..., bool], *a: Any, **kw: Any) -> Expectation:
        return later(kf(pred, *a, **kw).true)

    @property
    def buffer(self) -> Buffer:
        return self.vim.buffer

    @property
    def _buffer_out(self) -> List[str]:
        return self.buffer.content

    @property
    def contentkf(self) -> Expectable:
        return kf(lambda: self._buffer_out)

    def _list_line(self, log: Callable[[], List[str]], index: int, matcher: BoundMatcher) -> Expectation:
        minlen = index if index >= 0 else abs(index + 1)
        return later(
            kf(log).must(have_length(greater(minlen))) &
            kf(lambda: log().lift(index)).must(matcher)
        )

    def _log_line(self, index: int, matcher: BoundMatcher) -> Expectation:
        return self._list_line(lambda: self._log_out, index, matcher)

    def _log_contains(self, target: Union[str, BoundMatcher[str]]) -> Expectation:
        matcher = target if isinstance(target, BoundMatcher) else equal(target)
        return later(kf(lambda: self._log_out).must(contain(matcher)))

    def _log_equals(self, lines: Union[str, List[str]]) -> Expectation:
        return later(kf(lambda: self._log_out).must(have_lines(lines)))

    def _buffer_content(self, data: Union[str, List[str]]) -> Expectation:
        return later(self.contentkf.must(have_lines(data)))

    def _buffer_length(self, length: int) -> Expectation:
        return later(self.contentkf.must(have_length(length)))

    def _buffer_line(self, index: int, matcher: BoundMatcher) -> Expectation:
        return self._list_line(lambda: self._buffer_out, index, be_just(matcher))

    def _buffer_contains(self, line: str) -> Expectation:
        return later(kf(lambda: self._buffer_out).must(contain(line)))

    def _buffer_option(self, name: str, value: Any) -> Expectation:
        return later(kf(lambda: self.buffer.options(name)).must(be_right(value)))

    def _buffer_name(self, name: str) -> Expectation:
        return later(kf(lambda: self.buffer.name) == name)

    def _buffer_count(self, count: int) -> Expectation:
        return later(kf(lambda: self.vim.buffers).must(have_length(1)))

    def _window_height(self, height: int) -> Expectation:
        return later(kf(lambda: self.vim.window.height).must(contain(height)))

    def _window_count(self, count: int) -> Expectation:
        return later(kf(lambda: self.vim.windows).must(have_length(count)))

    def _cursor_line(self, lnum: int) -> Expectation:
        return later(kf(lambda: self.vim.window.line).must(contain(lnum)))

    def _messages_contain(self, line: str) -> Expectation:
        return later(kf(lambda: self.vim.messages).must(contain(line)))

    def seen_message(self, tpe: Type[M], **kw) -> Expectation:
        return later(kf(self.message_log).must(be_right(contain(have_type(tpe)))), **kw)

    def seen_times(self, tpe: Type[M], count: int, **kw) -> Expectation:
        seen = lambda: self.message_log() / __.filter_type(tpe)
        return later(kf(seen).must(be_right(have_length(count))), **kw)

    def var_is(self, name: str, value: Any) -> Expectation:
        return kf(self.vim.vars, name).must(be_right(value))

    def var_becomes(self, name: str, value: Any) -> Expectation:
        return later(self.var_is(name, value))

    def pvar_is(self, name: str, value: Any) -> Expectation:
        return kf(self.vim.vars.p, name).must(be_right(value))

    def pvar_becomes(self, name: str, value: Any) -> Expectation:
        return later(self.pvar_is(name, value))

    def pvar_becomes_map(self, name: str, value: Any, f: Callable[[Any], Any]) -> Expectation:
        return later(kf(lambda: self.vim.vars.p(name).map(f)).must(be_right(value)))
        return self._wait_for(lambda: self.vim.vars.p(name).map(f).contains(value))

    def command_exists(self, name: str, **kw) -> Expectation:
        return later(kf(self.vim.command_exists, name).true, **kw)

    def command_exists_not(self, name: str, **kw: Any) -> Expectation:
        return later(kf(self.vim.command_exists, name).false, **kw)


class VimIntegrationKlkSpec(VimIntegrationSpec, VimIntegrationKlkHelpers):
    pass


class ExternalIntegrationKlkSpec(ExternalIntegrationSpec, VimIntegrationKlkHelpers):
    pass


class PluginIntegrationKlkSpec(PluginIntegrationSpec, VimIntegrationKlkHelpers):
    pass

__all__ = ('VimIntegrationKlkHelpers', 'VimIntegrationKlkSpec', 'ExternalIntegrationKlkSpec',
           'PluginIntegrationKlkSpec')
